// Author: Hugo Pointcheval
// Email: git@pcl.ovh
// -----
// File: app.dart
// Created Date: 19/08/2022 12:05:38
// Last Modified: Fri Aug 26 2022
// -----
// Copyright (c) 2022

import 'package:example_router/core/routes/router.dart';
import 'package:example_router/core/utils/forms.dart';
import 'package:example_router/presentation/features/home/home_page.dart';
import 'package:example_router/presentation/features/welcome/welcome_page.dart';
import 'package:flutter/material.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:go_router/go_router.dart';
import 'package:wyatt_authentication_bloc/wyatt_authentication_bloc.dart';

class App extends StatelessWidget {
  final AuthenticationRepository authenticationRepository =
      AuthenticationRepositoryFirebase();

  App({Key? key}) : super(key: key);

  @override
  Widget build(BuildContext context) {
    AuthenticationState? previous;

    final AuthenticationCubit authenticationCubit = AuthenticationCubit(
      authenticationRepository: authenticationRepository,
      onAuthSuccess: (user) async {
        debugPrint(user.toString());
        return {};
      },
    );

    final GoRouter router = GoRouter(
      initialLocation: '/',
      routes: AppRouter.routes,
      debugLogDiagnostics: true,
      errorBuilder: (_, __) => const ColoredBox(
        color: Colors.red,
      ),
      refreshListenable: GoRouterRefreshStream(authenticationCubit.stream),
      redirect: (state) {
        final authState = authenticationCubit.state;

        if (authState != previous) {
          previous = authState;
          // Check if current user is logged in
          final loggedIn =
              authState.status == AuthenticationStatus.authenticated;

          // Checking if current path is onboarding or not
          final isOnboarding = AppRouter.publicRoutes.contains(state.subloc);

          if (!loggedIn) {
            debugPrint('Not logged');
            if (isOnboarding) {
              return null;
            } else {
              return state.namedLocation(WelcomePage.pageName);
            }
          } else {
            final email = authState.user?.email;
            debugPrint('Logged as: $email');
            if (isOnboarding) {
              return state.namedLocation(HomePage.pageName);
            } else {
              return null;
            }
          }
        }
        return null;
      },
    );

    return MultiRepositoryProvider(
      providers: [
        RepositoryProvider<AuthenticationRepository>.value(
          value: authenticationRepository,
        ),
      ],
      child: MultiBlocProvider(
        providers: [
          BlocProvider<AuthenticationCubit>.value(
            value: authenticationCubit..init(),
          ),
          BlocProvider<SignUpCubit>(
            create: (_) => SignUpCubit(
              authenticationRepository: authenticationRepository,
              formData: Forms.getNormalData(),
              onSignUpSuccess: (state, uid) async {
                debugPrint(state.toString());
                debugPrint(uid);
              },
            ),
          ),
          BlocProvider<SignInCubit>(
            create: (_) => SignInCubit(
              authenticationRepository: authenticationRepository,
            ),
          ),
        ],
        child: MaterialApp.router(
          title: 'Demo Authentication',
          debugShowCheckedModeBanner: false,
          routerDelegate: router.routerDelegate,
          routeInformationParser: router.routeInformationParser,
          routeInformationProvider: router.routeInformationProvider,
        ),
      ),
    );
  }
}
