// Copyright (C) 2022 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'dart:async';

import 'package:equatable/equatable.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:wyatt_authentication_bloc/src/core/enum/auth_cubit_status.dart';
import 'package:wyatt_authentication_bloc/src/domain/entities/user.dart';
import 'package:wyatt_authentication_bloc/src/domain/repositories/authentication_repository.dart';

part 'authentication_state.dart';

class AuthenticationCubit<Extra> extends Cubit<AuthenticationState<Extra>> {
  final AuthenticationRepository _authenticationRepository;
  late final StreamSubscription<AuthCubitStatus> _statusSubscription;

  StreamSubscription<User>? _userSubscription;

  final Future<Extra?> Function(User user)? _onAuthSuccess;

  AuthenticationCubit({
    required AuthenticationRepository authenticationRepository,
    Future<Extra?> Function(User user)? onAuthSuccess,
  })  : _authenticationRepository = authenticationRepository,
        _onAuthSuccess = onAuthSuccess,
        super(const AuthenticationState.unknown()) {
    _subscribeStatus();
  }

  Future<AuthCubitStatus> get status async =>
      _authenticationRepository.cubitStatus.last;

  void _subscribeStatus() {
    _statusSubscription = _authenticationRepository.cubitStatus.listen(
      (status) {
        switch (status) {
          case AuthCubitStatus.started:
            start();
            break;
          case AuthCubitStatus.stoped:
            stop();
            break;
        }
      },
    );
  }

  Future<void> init() async {
    final firstUser = await _authenticationRepository.user.first;
    _authenticationRepository.changeCubitStatus(AuthCubitStatus.started);
    return changeStatus(firstUser);
  }

  bool start() {
    _userSubscription = _authenticationRepository.user.listen(changeStatus);
    return true;
  }

  bool stop() {
    _userSubscription?.cancel();
    return true;
  }

  Future<void> changeStatus(User user) async {
    if (user.isNotEmpty) {
      final Extra? extra = await _onAuthSuccess?.call(user);
      emit(AuthenticationState.authenticated(user, extra));
    } else {
      _authenticationRepository.changeCubitStatus(AuthCubitStatus.stoped);
      emit(const AuthenticationState.unauthenticated());
    }
  }

  void logOut() {
    unawaited(_authenticationRepository.signOut());
  }

  @override
  Future<void> close() {
    _userSubscription?.cancel();
    _statusSubscription.cancel();
    return super.close();
  }
}
